package ittp;

import java.net.*;
import java.io.*;
import ittp.protocol.parser.RequestParser;
import ittp.protocol.request.Request;
import ittp.protocol.request.headers.Connection;
import ittp.protocol.response.Response;
import ittp.protocol.Protocol;
import gnu.getopt.LongOpt;
import gnu.getopt.Getopt;
import org.apache.commons.io.CopyUtils;
import java.util.HashMap;
import java.text.FieldPosition;
import java.io.DataInputStream;
import java.rmi.*;
import java.rmi.server.*;
import java.security.PermissionCollection;
import java.security.Permissions;
import java.nio.channels.SocketChannel;

/**
 * <B>The main entry to the ITTP Server </B>
 * Create the different a new socket for each request from a new client.
 * @author Frederic Bidon and Mathieu Texier
 * @version 1.3
 * @see Control
 */
public class Server extends Thread {
    /**
     * Current version of the server
     */
    final static public double VERSION = 1.0;
    /**
     * the whole connect thread
     */
    ThreadGroup threadGroup = new ThreadGroup ("ThreadPool");
    static private Server server = new Server ();
    private Parameters parameters;
    private PrintStream err;
    private PrintStream out;
    private InetAddress ip;
    private String rmiServer;
    /**
     * Constructor of the server.
     * Initialize the different parameters of the server
     */
    private Server () {
        super ("ServerThread");
        System.setSecurityManager (new RMISecurityManager ());
        try {
            ip = InetAddress.getLocalHost ();
            parameters = new Parameters ();
        } catch (IOException e) {
            System.err.println (e.getMessage ());
        }
    }
    
    /**
     * accessor to the object parameter
     * @return the object paramters that contains the different parameters of the server.
     */
    public static Parameters getParameters () {
        return server.parameters;
    }
    
    /**
     * Accesor for the one and only one server instance
     * @return the object server.
     */
    public static Server getInstance () {
        return server;
    }
    
    /**
     * Accesor for the one and only one server instance
     * @return the object server.
     */
    public static InetAddress getIP () {
        return server.ip;
    }
    
    /**
     * Method explaining to the user the operating of the several options.
     */
    
    public static void usage (){
        String Usage = "ITTP Server v" + VERSION + "\r\n" +
        "Usage: java Server [options]+\r\n\r\n" +
        "\t-p <port>\t\t--port <port>\r\n" +
        "<port> : Value of the port listen by the socket server.\r\n\r\n"+
        "\t-r <rmiServer>\t\t--rmiServer <rmiServer>\r\n" +
        "<rmiServer> : Adress where rmiregistry is executed.\r\n\r\n"+
        "\t-b <base_dir>\t\t--base_dir <base_dir>\r\n" +
        "<base_dir> : Directory where the client can request a resource.\r\n\r\n"+
        "\t-l <filelogerr>\t\t--ErrLog <filelogerr>\r\n" +
        "<filelogerr> : Name of the file where the errors are stored.\r\n\r\n"+
        "\t-i <fileloginfo>\t--InfoLog <fileloginfo>\r\n" +
        "<fileloginfo> Name of the file where different information of\n the server are stored.\r\n\r\n"+
        "\t-h\t\t\t--help\r\n" +
        "This option prints a synopsis of standard option.\r\n\r\n";
        Usage += "\tConcurrent Programming : course 02220\r\n" +
        "\tAuthor: Frederic Bidon, Texier Mathieu\r\n" +
        "\tTechnical University of Denmark\r\n";
        System.out.println (Usage);
    }
    
    /**
     * The main method of the server.
     * Check and initialize a new port's number or a new base directory if it are precised in parameters.
     * Starts a new ITTP Server listening on the specified port.
     * @param args the command line arguments
     */
    public static void main (String[] args) {
        
        LongOpt[] longopts = new LongOpt[6];
        StringBuffer sb = new StringBuffer ();
        longopts[0] = new LongOpt ("port", LongOpt.REQUIRED_ARGUMENT, sb, 'p');
        longopts[1] = new LongOpt ("rmi", LongOpt.REQUIRED_ARGUMENT, sb, 'r');
        longopts[2] = new LongOpt ("base_dir", LongOpt.REQUIRED_ARGUMENT, sb, 'b');
        longopts[3] = new LongOpt ("ErrLog", LongOpt.REQUIRED_ARGUMENT, sb, 'l');
        longopts[4] = new LongOpt ("InfoLog", LongOpt.REQUIRED_ARGUMENT, sb, 'i');
        longopts[5] = new LongOpt ("help", LongOpt.NO_ARGUMENT, null, 'h');
        
        Getopt g = new Getopt ("Server ITTP V" + VERSION, args, "p:r:b:l:i:h", longopts);
        g.setOpterr (false);
        
        server.rmiServer = "rmi://localhost/IttpServer";
        String base_dir = System.getProperty ("user.dir");
        String logE = "logErr.log";
        String logI = "logInfo.log";
        String port = "36000";
        
        BufferedReader in = new BufferedReader (new InputStreamReader (System.in));
        int c;
        while ((c = g.getopt ()) != -1) {
            try {
                switch (c) {
                    case 'p' : // number of port
                        port = g.getOptarg ();
                        if (! port.matches ("^0*[1-9]+0*"))
                            throw new IllegalArgumentException ("Error with port parameter.");
                        if ((Integer.parseInt (port) < 1) || (Integer.parseInt (port) > 65535))
                            throw new IllegalArgumentException ("Error with port parameter.");
                        
                        break;
                    case 'r' : // base directory
                        server.rmiServer = g.getOptarg ();
                        break;
                    case 'b' : // base directory
                        base_dir = g.getOptarg ();
                        File f =  new File (base_dir);
                        if ((!f.exists ())||(!f.isDirectory ())||(!f.isAbsolute ()))
                            throw new IllegalArgumentException ("Error with base_directory parameter.");
                        break;
                    case 'l' : // file log Errors
                        logE = g.getOptarg ();
                        File flogE =  new File (logE);
                        if (!flogE.isFile ())
                            throw new IllegalArgumentException ("Error with log parameter.");
                        break;
                    case 'i' : // file log Info
                        logI = g.getOptarg ();
                        File flogI =  new File (logE);
                        if (!flogI.isFile ())
                            throw new IllegalArgumentException ("Error with log parameter.");
                        break;
                    case 'h' : // help
                        System.out.println ("help");
                        usage ();
                        System.exit (0);
                        break;
                    default:
                        throw new IllegalArgumentException ("Invalid command.");
                }
            } catch (IllegalArgumentException e) {
                System.err.println (e.getMessage ());
                System.err.println ("Execute the program with -h for help.");
                System.exit (1);
            }
        }
        for (int i = g.getOptind (); i < args.length ; i++) {
            System.err.println ("Nonvalid argument: " + args[i] + "\n");
            usage ();
            System.exit (1);
        }
        
        try {
            server.err = new PrintStream (new FileOutputStream (new File (logE), true));
            server.out = new PrintStream (new FileOutputStream (new File (logI), true));
            getParameters ().setBaseDir (base_dir);
            getParameters ().setPort ((Integer.parseInt (port)));
            System.out.println ();
            String OptionValue =  "---- Parameters----\r\n" +
            " port :\t\t\t" + port + "\r\n"+
            " rmiServer :\t\t\t" + server.rmiServer + "\r\n"+
            " base directory:\t" + base_dir + "\r\n" +
            " error log file:\t" + logE + "\r\n" +
            " output log file:\t" + logI + "\r\n\r\n";
            
            System.out.println (OptionValue);
            
            server.start ();
            server.join ();
        } catch (IOException e) {
            System.err.println (e.getMessage ());
        } catch (InterruptedException e) {;}
    }
    
    /**
     * Accepts new client connections.
     * This method is called by a new thread when the superclass
     * call the start() method.
     */
    public void run () {
        ServerSocket s = null;
        try {
            Naming.rebind (rmiServer, parameters);
        } catch (IOException e) {
            parameters.recordError ();
            logInfo (err, e.getMessage());
            System.exit (1);
        }
        try {
            s = new ServerSocket (parameters.getPort ());
        } catch (IOException e) {
            parameters.recordError ();
            logInfo (err, "the serverSocket has not been instantiated. May be the Server is allready running");
            System.exit (1);
        }
        
        String currentState = Parameters.STOP;
        boolean refuse = false;
        boolean exit = false;
        while (!exit) {
            try {
                String currentStateTest = parameters.getServerState ();
                if (currentStateTest != currentState)
                    logInfo (out, "Server " + currentStateTest);
                currentState = currentStateTest;
                
                if (parameters.getNumberOfActiveConection () < parameters.getConnectionMax ()
                && currentState.equals (Parameters.START))
                    refuse = false;
                else
                    refuse = true;
                
                if (currentState.equals (Parameters.SHUT_DOWN) && parameters.getNumberOfActiveConection () == 0) {
                    System.exit (0);
                }
            } catch (IOException ex) {
                logInfo (err, ex.getMessage ());
                parameters.recordError ();
            }
            
            if (!refuse) {
                try {
                    Connect connect = new Connect (s.accept ());
                    connect.start ();
                } catch (IOException e) {logInfo (err, e.getMessage ());}
            } else {
                try {
                    Socket sck = s.accept ();
                    PrintWriter sckOut = new PrintWriter (sck.getOutputStream (), true);
                    Response response = new Response ();
                    response.code.setResponseCode (Response.Code.ITTP_UNAVAILABLE);
                    sckOut.flush ();
                    logInfo (err, "Service unavailable");
                    sckOut.close ();
                    sck.close ();
                } catch (IOException e) {logInfo (err, e.getMessage ());}
            }
        }
    }
    
    /**
     * Redirect the different messages generated during the execution of
     * the different thread to different file according to if we have an error's message or an information's message.
     * @param message String containing the error message denerated by the server.
     * @param std Stream into witch the message is written
     */
    static public synchronized void logInfo (PrintStream std, String message) {
        StringBuffer strDate = new StringBuffer ();
        strDate = Protocol.dateFormat.format (new java.util.Date (), strDate ,new FieldPosition (0));
        std.println (strDate + " " + message);
        System.out.println (strDate + " " + message);
    }
    
    /**
     * This class can be executed by several thread.
     * Each thread manage a connection with a client.
     * We have one thread by client.
     */
    private class Connect extends Thread {
        private Socket sck;
        private boolean broken = false;
        private OutputStream sckOut = null;
        private BufferedReader sckIn = null;
        private int timeOut = 10; //[s]
        final static int frequency = 4; //[hz]
        InetAddress clientIP;
        
        /**
         * Constructor of the class.
         * Read the contains of the socket.
         * @param sck: Socket used to communicate between the client and the server.
         */
        public Connect (Socket sck) {
            super (threadGroup, (Runnable) null);
            try {
                timeOut = parameters.getTimeOut ();
                sckIn = new BufferedReader (new InputStreamReader (sck.getInputStream ()));
                sckOut = sck.getOutputStream ();
            } catch (IOException e) {
                logInfo (err, e.getMessage ());
            }
            this.sck =sck;
            this.clientIP = sck.getInetAddress ();
            logInfo (out, "Accepted from " + sck.getInetAddress ());
            parameters.recordConnect ();
        }
        
        /**Execution of the threads.*/
        public void run () {
            do {
                Response response = new Response ();
                
                try {   // Read body every second until the timeout
                    int counter = 0;
                    while (!sckIn.ready () && (!broken)) {
                        sleep (1000/frequency);
                        if (++ counter > timeOut*frequency)
                            broken = true;
                    }
                } catch (IOException e) {
                    logInfo (err, e.getMessage ());
                    broken = true;
                } catch (InterruptedException e) {
                    logInfo (err, e.getMessage ());
                    broken = true;
                }
                
                if (!broken)
                    try {
                        parameters.recordRequest ();
                        RequestParser p = new RequestParser (sckIn, clientIP);
                        Request request = p.parse ();
                        request.process (response);
                        
                        Connection connection = (Connection) request.getHeaders (request.CONNECTION);
                        boolean closeAfter = (connection.getConnection () == Connection.CLOSE);
                        send (response);
                        if (closeAfter)
                            broken = true;
                        
                    } catch (ittpException e) {
                        logInfo (err, e.getMessage ());
                        int r = e.getResponseCode ();
                        response.code.setResponseCode (r);
                        try {
                            send (response);
                        } catch (IOException er) {
                            logInfo (err, e.getMessage ());
                        }
                        broken = true;
                    } catch (IOException e) {
                        logInfo (err, e.getMessage ());
                    }
            } while (!broken);
            close ();
        }
        
        /**
         * Send the response to the client
         * @param response: Response returned to the client by the socket.
         * @throws a IOException
         */
        private void send (Response response) throws IOException {
            sckOut.write ('\n');
            CopyUtils.copy (response.toStream (), sckOut);
            sckOut.write ('\n');
            if (response.body != null)
                parameters.recordOctet (response.body.length ());
            out.flush ();
        }
        
        /**
         * Close the socket, the BufferedReader and the OutputStream if the socket is instantiated.
         */
        private void close () {
            logInfo (out, "close" + sck);
            try {
                if (sckIn != null)
                    sckIn.close ();
                if (sckOut != null)
                    sckOut.close ();
                if (sck != null)
                    sck.close ();
            } catch (Exception e){
                logInfo (err, "Error finish execution thread " + e.getMessage ());
            }
        }
    }
}
