package ittp.protocol.parser;

import ittp.protocol.request.*;
import ittp.protocol.request.headers.*;
import java.io.BufferedReader;
import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.net.Socket;

/**
 * The aim of this class is to parse the request send by the client.
 * @author Frederic Bidon and Mathieu Texier
 */
public class RequestParser {
    
    private TokenStream ts;
    private Request request;
    private InetAddress clientIP;
    
    /**Create an instance of the RequestParse
     * we assume the bufferReader is ready
     * param in InputStream to be parse
     * param clientIP address of the client
     */
    public RequestParser (BufferedReader in,InetAddress clientIP) {
        ts = new ITTPScan (in);
        this.clientIP = clientIP;
    }
    
    /**Execute the method that analyse the request of the client.*/
    private void request () {
        request_line ();
        requireNL ();
        while (ts.tok != TokenStream.NL && ts.tok != TokenStream.EOF)
            request_head ();
        requireNL ();
    }
    
    /**This method analyse the headers present in the request head.*/
    private void request_head () {
        switch (ts.tok) {
            case TokenStream.CONNECTION:
                ts.next ();
                requireColon ();
                if (ts.tok == TokenStream.WORD && ts.sval.equals ("close")) {
                    ((Connection) request.getHeaders (Request.CONNECTION)).setConnection (Connection.CLOSE);
                    ts.next ();
                } else throw ts.ParseException ("Expected \"close\" ");
                break;
            case TokenStream.DATE:
                ts.next ();
                requireColon ();
                ((Date) request.getHeaders (Request.DATE)).setDateTime (ts.next (6));
                ts.next ();
                break;
            case TokenStream.HOST:
                ts.next ();
                if (ts.tok  != TokenStream.COLON)
                    throw ts.ParseException ("Expected ':'");
                ts.filterString ();
                ts.next ();
                if (ts.tok == TokenStream.WORD) {
                    String host = ts.sval;
                    ((Host) request.getHeaders (Request.HOST)).setHostName (host);
                } else throw ts.ParseException ("Expected host or IP ");
                ts.setup ();
                ts.next ();
                break;
            case TokenStream.IF_MODIFIED_SINCE:
                ts.next ();
                requireColon ();
                ((ModifiedSince) request.getHeaders (Request.MODIFIED_SINCE)).setModifiedSince (ts.next (6));
                ts.next ();
                break;
            case TokenStream.RANGE:
                ts.next ();
                requireColon ();
                int first = 0;
                int last = 0;
                if (ts.tok == TokenStream.NUMBER) {
                    first = (int) ts.nval;
                } else throw ts.ParseException ("Expected \"first\" ");
                ts.next ();
                
                if (ts.tok == TokenStream.NUMBER && ts.nval <0) {
                    last = (int) Math.abs (ts.nval);
                }
                else if (ts.tok == TokenStream.MINUS) {
                    ts.next ();
                    if (ts.tok == TokenStream.NUMBER) {
                        last = (int) ts.nval;
                    }
                }
                else throw ts.ParseException ("Expected \"last\" ");
                ((Range) request.getHeaders (Request.RANGE)).setRange (first, last);
                ts.next ();
                break;
            default:
                break;
        }
        requireNL ();
    }
    
    private void request_line () {
        method ();
        request_URI ();
        ITTP_version ();
    }
    
    /**This method instantiate the different headers present in the request.*/
    private void method () {
        switch (ts.tok) {
            case TokenStream.GET:
                request = new Get ();
                break;
            case TokenStream.HEAD:
                request = new Head ();
                break;
            case TokenStream.OPTIONS:
                request = new Option ();
                break;
            case TokenStream.TRACE:
                request = new Trace ();
                break;
            default:
                throw ts.ParseException ("Unexpected method ");
        }
        request.setIP (clientIP);
        ts.next ();
    }
    
    void request_URI () {
        sheme ();
        requireSlash ();
        host ();
        path ();
    }
    /**This method analyse the sheme.*/
    void sheme () {
        switch (ts.tok) {
            case TokenStream.ITTP :
                request.setSheme (ts.sval);
                ts.next ();
                requireColon ();
                return;
            default:
                return;
        }
    }
    
    /**
     * This method analyse the host of the client.
     * It throws an exception if the host is not expected.
     */
    void host () {
        switch (ts.tok) {
            case TokenStream.SLASH :
                ts.filterString ();
                ts.next ();
                if (ts.tok == TokenStream.WORD) {
                    String host = ts.sval;
                    request.setHost (host);
                } else throw ts.ParseException ("Expected host or IP ");
                ts.setup ();
                ts.next ();
                port ();
                requireSlash ();
                return;
            default:
                return;
        }
    }
    
    /**
     * This method analyse the port required by the client.
     * It throws an exception if the port is not expected.
     */
    void port () {
        switch (ts.tok) {
            case TokenStream.COLON :
                if (ts.next () == TokenStream.NUMBER)
                    request.setPort ((int) ts.nval);
                else throw ts.ParseException ("Expected port ");
                ts.next ();
                return;
            default:
                return;
        }
    }
    
    /**
     * This method analyse the path of the ressource contained in the request.
     * It throws an exception if the path is not expected.
     */
    void path () {
        String path = new String ();
        while (true) {
            switch (ts.tok) {
                case TokenStream.WORD :
                    path += ts.sval;
                    ts.next ();
                    if (ts.tok == TokenStream.SLASH) {
                        ts.next ();
                        path += "/";
                    } else {
                        request.setPath (path);
                        return;
                    }
                    break;
                default: throw ts.ParseException ("Expected path ");
            }
        }
    }
    
    
    /**
     * This method analyse the ITTP version used by the client.
     * It throw an exception if the version is not expected.
     */
    void ITTP_version () {
        switch (ts.tok) {
            case TokenStream.WORD :
                String ITTP_version = ts.sval;
                ts.next ();
                requireSlash ();
                if (ts.tok == TokenStream.NUMBER) {
                    ITTP_version += "/" +  ts.nval;
                    request.setITTP_version (ITTP_version);
                    ts.next ();
                    return;
                }
                // deliberate Fallthrough
            default: throw ts.ParseException ("Expected ITTP/verion ");
        }
    }
    
    /**This method throw an exception when a new line is required, else it go to the next token.*/
    private void requireNL () {
        if (ts.tok != TokenStream.NL)
            throw ts.ParseException ("Expected end of line ");
        ts.next ();
    }
    
    /**This method throw an exception when a slash is required, else it go to the next token.*/
    private void requireSlash () {
        if (ts.tok != TokenStream.SLASH)
            throw ts.ParseException ("Expected '/' ");
        ts.next ();
    }
    
    /**This method throw an exception when a colon is required, else it go to the next token.*/
    private void requireColon () {
        if (ts.tok  != TokenStream.COLON)
            throw ts.ParseException ("Expected ':'");
        ts.next ();
    }
    
    /**This method throw an exception when a coma is required, else it go to the next token.*/
    private void requireComa () {
        if (ts.tok  != TokenStream.COMA)
            throw ts.ParseException ("Expected ','");
        ts.next ();
    }
    
    /**
     * @return the parsed request.
     */
    public Request parse () {
        request ();
        return request;
    }
}
