package ittp.protocol.response;

import ittp.RemoteControl;
import ittp.protocol.Protocol;
import ittp.ittpException;
import ittp.protocol.request.Request;
import ittp.protocol.response.headers.*;
import org.apache.soap.encoding.soapenc.Base64;

import java.util.HashMap;
import java.util.Iterator;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.SequenceInputStream;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.rmi.RemoteException;

/**
 * This class build the response send to the client according to the method and the headers.
 * @author Frederic Bidon and Mathieu Texier
 */
public class Response extends Protocol {
    public Code code = new Code ();
    public Headers headers;
    public Body body = new Body ();
    private HashMap headerList;
    /**Declaration of the different headers send in a response*/
    static final public String ALLOW = "Allow";
    static final public String CONNECTION = "Connection";
    static final public String CONTENT_LENGTH = "Content-Length";
    static final public String CONTENT_MD5 = "Content-MD5";
    static final public String CONTENT_RANGE = "Content-Range";
    static final public String CONTENT_TYPE = "Content-type";
    static final public String DATE = "Date";
    static final public String HOST = "Host";
    static final public String LAST_MODIFIED = "Last-Modified";
    static final public String SERVER = "Server";
    
    private boolean virtualBody = false;
    
    /** Creates a new instance of Headers */
    public Response () {
        headerList = new HashMap ();
        headerList.put (ALLOW, new Allow (this));
        headerList.put (CONNECTION, new Connection (this));
        headerList.put (CONTENT_LENGTH, new ContentLength (this));
        headerList.put (CONTENT_MD5, new ContentMD5 (this));
        headerList.put (CONTENT_RANGE, new ContentRange (this));
        headerList.put (CONTENT_TYPE, new ContentType (this));
        headerList.put (DATE, new Date (this));
        headerList.put (LAST_MODIFIED, new LastModified (this));
        headerList.put (SERVER, new Server (this));
        
        getHeaders (Response.CONNECTION).setAllowed (true);
        getHeaders (Response.CONTENT_LENGTH).setAllowed (true);
        getHeaders (Response.CONTENT_RANGE).setAllowed (true);
        getHeaders (Response.DATE).setAllowed (true);
        getHeaders (Response.LAST_MODIFIED).setAllowed (true);
        try {
            boolean isAllowed = ittp.Server.getParameters ().isAllowedHeader (Response.ALLOW);
            getHeaders (Response.ALLOW).setAllowed (isAllowed);
            
            isAllowed = ittp.Server.getParameters ().isAllowedHeader (Response.CONTENT_MD5);
            getHeaders (Response.CONTENT_MD5).setAllowed (isAllowed);
            
            isAllowed = ittp.Server.getParameters ().isAllowedHeader (Response.CONTENT_TYPE);
            getHeaders (Response.CONTENT_TYPE).setAllowed (isAllowed);
            
            isAllowed = ittp.Server.getParameters ().isAllowedHeader (Response.SERVER);
            getHeaders (Response.SERVER).setAllowed (isAllowed);
        } catch (RemoteException e) {
            throw new ittpException (e.getMessage (), Response.Code.ITTP_INTERNAL_ERROR);
        }
    }
    
    /**
     * Accessor for the object Headers
     * @param headerString String containing the name of the header.
     * @return an object Headers correponding to the name headerstring.
     */
    public Headers getHeaders (String headerString ) {
        return (Headers) headerList.get (headerString);
    }
    
    /**
     * Method to convert the response to a string.
     * The response can contain the body and the headers or just the headers.
     * @return the String retString containing the response of the server.
     */
    public String toString () {
        Iterator it = headerList.keySet ().iterator ();
        String retString = new String ("\n");
        retString = "ITTP/" + VERSION + " ";
        retString += code + "\n";
        if (code.getResponseCode () >= 400)
            return retString;
        while (it.hasNext ()) {
            String type = (String) it.next ();
            Headers h = (Headers) headerList.get (type);
            retString += (h.used && h.allowed) ? (type + ":" + h) : "";
        }
        return retString;
    }
    
    /**
     * Method to convert the headers and the body to an InputStream.
     * This stream can contain the stream of the body and the headers or just the stream of the headers.
     * @return an InputStream containin the stream of the response
     */
    public InputStream toStream () {
        InputStream headersStream = new ByteArrayInputStream (this.toString ().getBytes ());
        if ((virtualBody)||(code.getResponseCode () >= 400))
            return headersStream;
        return new SequenceInputStream (headersStream, body.toStream ());
    }
    
    /**
     * Mutators for the boolean virtualBody
     */
    public void setVirtualBody (boolean virtualBody) {
        this.virtualBody = virtualBody;
    }
    
    /**
     * This class create the body of the response according to the request of the client.
     */
    public class Body {
        public File file;
        private long length = 0;
        private int first = 0;
        private int last = 0;
        private String pathName = new String ();
        private byte[] content;
        private String digest = new String ();
        
        /**Constructor of the class*/
        private Body () {
            
        }
        
        /**
         * Constructor of the class for the whole of the file.
         */
        public Body (File file) {
            this (file, 0, (int)file.length ());
        }
        
        /**
         * Constructor of the class for a partial of the file.
         */
        public Body (File file, int first, int last) {
            body = this;
            this.file = file;
            this.first = first;
            this.last = last;
            this.pathName = file.getPath ();
            try {
                FileInputStream fileInputStream = new FileInputStream (file);
                fileInputStream.skip (first);
                content = new byte[last - first];
                length = fileInputStream.read (content,0, content.length);
            } catch (IOException e) {
                throw new ittpException ("Error in file " + pathName, Response.Code.ITTP_NOT_FOUND);
            }
            if ((first != 0)||(last != file.length ()))
                code.setResponseCode (Code.ITTP_PARTIAL);
            MessageDigest ();
        }
        
        /**
         * Constructor of the class for a trace method.
         */
        public Body (Request request){
            body = this;
            this.pathName = Request.methodName[Request.TRACE];
            content = request.toString ().getBytes ();
            length = content.length;
            MessageDigest ();
        }
        
        /**
         * This method allows to create the content-MD5 of the body.
         */
        private void MessageDigest () {
            try {
                MessageDigest messageDigest = MessageDigest.getInstance ("MD5");
                digest = Base64.encode (messageDigest.digest (content));
            } catch (NoSuchAlgorithmException e) {
                System.err.println (e);
                throw new ittpException ("Invalid algorithm", Response.Code.ITTP_INTERNAL_ERROR);
            }
        }
        
        /**
         * Accessors for the value of the content MD5.
         * @return the String digest containing the content-MD5 of the body.
         */
        public String getDigest () {
            return digest;
        }
        
        /**
         * Accessors for the length of the body.
         * @return the length of the body.
         */
        public long length (){
            return length;
        }
        
        /**
         * Accessors for the number of the first octet of the body.
         * @return the int first.
         */
        public int getFirst () {
            return first;
        }
        
        /**
         * Accessors for the number of the last octet of the body.
         * @return the int last.
         */
        public int getLast () {
            return last;
        }
        
        /**
         * Accessors for the name of the file (pathName).
         * @return the name of the file.
         */
        public String getPathName () {
            return pathName;
        }
        
        /**
         * Convert to an InputStream the content of the body
         * @return an InputStream containing the content of the program.
         */
        public InputStream toStream () {
            if (content == null)
                return null;
            else
                return new ByteArrayInputStream (content);
        }
    }
    
    /**
     * This class allow to know the state of the response.
     * It allows also to declare the different number of response code.
     */
    public class Code extends HashMap {
	
	/*
	 *  ITTP Status-Code 200: OK.
	 */        
        static final public int ITTP_OK = 200;
        
        /*
	 *  ITTP Status-Code 206: Partial Content.
	 */
        static final public int ITTP_PARTIAL = 206;
        
        /*
	 *  ITTP Status-Code 304: Not Modified.
	 */
        static final public int ITTP_NOT_MODIFIED = 304;
        
        /*
	 *  ITTP Status-Code 400: Bad Request.
	 */
        static final public int ITTP_BAD_REQUEST = 400;
        
        /*
	 *  ITTP Status-Code 404: Not Found.
	 */
        static final public int ITTP_NOT_FOUND = 404;
        
        /*
	 *  ITTP Status-Code 405: Method Not Allowed.
	 */
        static final public int ITTP_METHOD_NOT_ALLOWED = 405;
        
        /*
	 *  ITTP Status-Code 416: Requested Range Not Satisfiable.
	 */
        static final public int ITTP_OUT_OF_RANGE = 416;
        
        /*
	 *  ITTP Status-Code 500: Internal Server Error.
	 */
        static final public int ITTP_INTERNAL_ERROR = 500;
        
        /*
	 *  ITTP Status-Code 501: Not Implemented.
	 */
        static final public int ITTP_NOT_IMPLEMENTED = 501;
        
        /*
	 *  ITTP Status-Code 503: Service Unavailable.
	 */
        static final public int ITTP_UNAVAILABLE = 503;
        
        /*
	 *  ITTP Status-Code 505: HTTP Version Not Supported.
	 */
        static final public int ITTP_VERSION = 505;
        
        private int response_code = Code.ITTP_OK;
        
        /**
         * Constructor of the class.
         * It associate each code to a corresponding error message.
         */
        public Code () {
            put (new Integer (ITTP_OK), "OK");
            put (new Integer (ITTP_PARTIAL), "Partial Content.");
            put (new Integer (ITTP_NOT_MODIFIED), "Not Modified. ");
            put (new Integer (ITTP_BAD_REQUEST), "Syntax Error In Request.");
            put (new Integer (ITTP_NOT_FOUND), "Ressource Not Found.");
            put (new Integer (ITTP_METHOD_NOT_ALLOWED), "Method Not Allowed.");
            put (new Integer (ITTP_OUT_OF_RANGE), "Requested Range Not Satisfiable");
            put (new Integer (ITTP_INTERNAL_ERROR), "Internal Server Error.");
            put (new Integer (ITTP_NOT_IMPLEMENTED), "Method Not Implemented.");
            put (new Integer (ITTP_UNAVAILABLE), "Service Unavailable.");
            put (new Integer (ITTP_VERSION), "ITTP Version Not Supported.");
        }
        
        /**
         * Mutators for the response_code.
         * If the code is superior to 400, then it is register like an error.
         * @param response_code int containing the number of the error code.
         */
        public void setResponseCode (int response_code) {
            this.response_code = response_code;
            if (response_code >= 400)
                ittp.Server.getParameters ().recordError ();
        }
        
        /**
         * Accessors that allows to return the code of the state of the response.
         * @return the response code.
         */
        int getResponseCode () {
            return response_code;
        }
        
        /**
         * Accessors that allows to return the code and the message of the state of the response.
         * @return the error message that the clieant will receive.
         */
        public String toString () {
            return Integer.toString (response_code) +" "+ (String) get ( new Integer (response_code));
        }
    }
    
    /**
     * Abstract class headers that allows to build the different headers for the response.
     */
    abstract public class Headers {
        protected boolean used = true;
        protected boolean allowed = true;
        abstract public void setUsed (boolean used);
        abstract public void setAllowed (boolean allowed);
        abstract public String toString ();
    }
}
